#!/usr/bin/python3

#
#    binToHuman.py --- Part of khoca, a knot homology calculator
#
# Copyright (C) 2018 Lukas Lewark <lukas@lewark.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# Reads from stdin, converts from the binary format to the human readable.

import sys

qShift_t = -16
tShift_t = qShift_t
dot_t = 16
vidx_t = 32
vidx2_t = 64
boundary_t = -8
uint32_t = 32

def myRead(size):
	assert(size % 8 == 0)
	return int.from_bytes(sys.stdin.buffer.read(int(abs(size / 8))), byteorder='little', signed=(size < 0))


def krasnerTangle(boundarySize, indent):
	print("  " * indent + "Tangle: Q-shift = " + str(myRead(qShift_t)) + " + " + str(myRead(qShift_t)) + "*N, " + str(myRead(boundary_t)) + " circles.")
	print("  " * (indent + 1) + "Data: " + " ".join([str(myRead(boundary_t)) for i in range(boundarySize)]))

def vecTangles(boundarySize, indent):
	tanglesSize = myRead(vidx2_t)
	print("  " * indent + "Module: Sum of " + str(tanglesSize) + " tangles.")
	for i in range(tanglesSize):
		krasnerTangle(boundarySize, indent + 2)
	deloopStackSize = myRead(vidx2_t)
	print("  " * (indent + 1) + "The deloop-stack has size " + str(deloopStackSize) + " and consists of the elements:" + " ".join([str(myRead(vidx2_t)) for i in range(deloopStackSize)]))

def mpirUInt():
	size = myRead(8)
	if (size == 0):
		return "0"
	else:
		return str(myRead(size * 8))

def mpirRat(coefficientRing):
	if (coefficientRing == 0):
		sgn = myRead(8)
		if (sgn == 0):
			return str(myRead(32))
		else:
			return str(-myRead(32))
	elif (coefficientRing == 1):
		sgn = myRead(8)
		num = mpirUInt()
		den = mpirUInt()
		if (sgn != 0):
			num = "-" + num
		return num + "/" + den
	elif (coefficientRing < 256):
		return str(myRead(8))
	else:
		return str(myRead(16))

def krasnerCobo(indent, coefficientRing):
	s = "  " * indent + mpirRat(coefficientRing) + " * cobordism with "
	size = myRead(boundary_t)
	s += str(size) + " facets and the following dot distribution: " + " ".join([str(myRead(dot_t)) for i in range(size)])
	print(s)

def LCCobos(indent, coefficientRing):
	size = myRead(vidx2_t)
	print("  " * indent + "LC of " + str(size) + " cobordisms:")
	for i in range(size):
		krasnerCobo(indent + 1, coefficientRing)

def matLCCobos(indent, coefficientRing):
	colCount = myRead(uint32_t)
	entryCount = myRead(uint32_t)
	print("  "*indent + "Map (as a matrix in yale format): " + str(colCount) + " columns, " + str(entryCount) + " entries.")
	print("  "*(indent + 1) + "Listing entries:")
	for i in range(entryCount):
		LCCobos(indent + 2, coefficientRing)
	print("  "*(indent + 1) + "Listing column indices:" + " ".join([str(myRead(uint32_t)) for i in range(entryCount)]))
	rowPtrSize = myRead(uint32_t)
	print("  "*(indent + 1) + "Number of rows plus one = " + str(rowPtrSize) + ", and the row pointers: " + " ".join([str(myRead(uint32_t)) for i in range(rowPtrSize)]))
	invertiblesSize = myRead(uint32_t)
	print("  "*(indent + 1) + "Number of invertible entries = " + str(invertiblesSize) + ", and their indices: " + " ".join([str(myRead(uint32_t)) for i in range(invertiblesSize)]))

def complex(indent):
	coefficientRing = myRead(tShift_t)
	print("Coefficient ring is " + str(coefficientRing))
	globalTShift = myRead(qShift_t)
	boundarySize = myRead(boundary_t)
	size = myRead(vidx_t)
	print("  "*indent + "Chain complex: global t-shift = " + str(globalTShift) + ", " + str(boundarySize) + " boundary points, " + str(size) + " modules.")
	print("  "*(indent + 1) + "Listing modules:")
	for i in range(size):
		vecTangles(boundarySize, indent + 2)
	print("  "*(indent + 1) + "Listing maps:")
	for i in range(size - 1):
		matLCCobos(indent + 2, coefficientRing)

complex(0)



#!/usr/bin/env python

# A packaging of Khoca that allows easy installation using python's pip
#
# Sebastian Oehms, 09/24/2021
# seb.oehms@gmail.com
#
# Run "python setup.py package" and it will automatically download all the
# necessary sources and create a tar ball suitable for pip.
#
# We can upload with "twine upload -r pypi khoca-...tar.gz"

import glob, os


# Without the next line, we get an error even though we never
# use distutils as a symbol
from setuptools import distutils

from distutils.core import Extension
from setuptools import setup
from Cython.Build import cythonize
from platform import system

join = os.path.join
env = os.environ

v_file = join(os.path.dirname(__file__), '_version.py')
version = '0.9'
with open(v_file) as f:
    # read the current version from file
    code = f.read()
    exec(code, locals())

khoca_dir      = '.'
src_dir        = 'src'
bin_dir        = 'bin'
data_dir       = 'data'
converters_dir = 'converters'

khoca_pkg      = 'khoca'
bin_pkg        = join(khoca_pkg, bin_dir)
data_pkg       = join(khoca_pkg, data_dir)
converters_pkg = join(khoca_pkg, converters_dir)

pui_name = 'khoca.bin.pui'

Linux = (system() == 'Linux')
MacOS = (system() == 'Darwin')
Windows = (system() == 'Windows')

include_dirs = []
library_dirs = []
extra_objects = []
extra_compile_args = ['-c', '-D__STDC_LIMIT_MACROS', '-Wall']
extra_link_args = ['-lpthread', '-lstdc++', '-t']
libraries = []

if Linux:
    extra_compile_args += ['-fopenmp', '-std=c++11', '-shared', '-fPIC', '-O3']
    extra_link_args += ['-z defs']
    libraries = ['gmp','gmpxx','pari']
elif MacOS:
    locdir = ('Pari42')
    pari_include_dir = join(locdir, 'include')
    pari_library_dir = join(locdir, 'lib')
    hombrew_lib = '/opt/homebrew/lib/'
    include_dirs += ['/opt/homebrew/opt/libomp/include', '/opt/homebrew/include/', pari_include_dir]
    library_dirs += ['/opt/homebrew/opt/libomp/lib', hombrew_lib, pari_library_dir]
    extra_compile_args += ['-std=c++11', '-shared', '-fPIC', '-O3', '-mmacosx-version-min=10.9', '-Wno-unreachable-code', '-Wno-unreachable-code-fallthrough']
    libraries = ['gmp','gmpxx', 'pari']
elif Windows:
    locdir = ('Pari42')
    pari_include_dir = join(locdir, 'include')
    pari_library_dir = join(locdir, 'bin')
    gmp_include_dir = r'C:\msys64\usr\include'
    gmp_library_dir = r'C:\msys64\mingw64\lib'
    gmp_library_dir_bin = r'C:\msys64\mingw64\bin'

    include_dirs += [pari_include_dir, gmp_include_dir]
    library_dirs += [gmp_library_dir, gmp_library_dir_bin, pari_library_dir]
    extra_compile_args += ['/DDISABLE_INLINE', '/openmp', '/std:c11', '/LD']
    extra_link_args = [join(gmp_library_dir, 'libgmp.dll.a'), join(gmp_library_dir, 'libgmpxx.dll.a'), join(pari_library_dir, 'libpari.dll.a')]

def local_scheme(version):
    return ""

def collect_source(path, pattern, depth=0):
    """
    Find all files with the given extension under path.
    """
    result = []
    for l in range(depth + 1):
        path_components = path.split('/') + l * ['*'] + [pattern]
        result += glob.glob(join(*path_components))
    return result

def create_extension(name, cpps, includes):
    return Extension(
    name,
    sources=cpps,
    language='c++',
    include_dirs=includes,
    library_dirs=library_dirs,
    extra_objects=extra_objects,
    extra_compile_args=extra_compile_args,
    libraries=libraries,
    extra_link_args=extra_link_args)

template_cpp_files = collect_source('', '*Templates.cpp', depth=2)
cpp_files_with_templ = collect_source('', '*.cpp', depth=2)
cpp_files = [cpp for cpp in cpp_files_with_templ if cpp not in template_cpp_files]

data_files = collect_source('data/', '*')

print('cpp_files', cpp_files)
print('data_files', data_files)

include_dirs += ['', 'python_interface', 'planar_algebra', 'krasner']

pui_ext = create_extension(pui_name, cpp_files, include_dirs)

setup(name = khoca_pkg,
      version=version,
      license='GPLv2+',
      zip_safe=False,
      packages=[khoca_pkg, bin_pkg, converters_pkg, data_pkg],
      package_dir={
          khoca_pkg:      khoca_dir,
          bin_pkg:        bin_dir,
          converters_pkg: converters_dir,
          data_pkg:       data_dir
      },
      ext_modules=[pui_ext]+cythonize('src/python_interface/pui.pyx'),
      package_data={khoca_pkg: data_files},
      include_package_data=True,
      install_requires=[]
)



#!/usr/bin/python3

#
#    pseudoBraidToKrasnerGauss.py --- Part of khoca, a knot homology calculator
#
# Copyright (C) 2018 Lukas Lewark <lukas@lewark.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# Accepted input: A list of signed integers, where i stands for
# (sigma_i)^2 and -i for (sigma_i)^(-2). The braid is taken to
# have an even number of strands, which are then connected as follows:
# On the left, 1 to 2, 3 to 4 and so on; on the right, 1 to the last, 2 to 3, 4 to 5 and so on.
# Output is a list of signed integers representing the Krasner Gauss
# encoding of knot (so the output may be fed to KrasnerGaussToMyPD.py).

import sys, re
sys.path.append('.')
import pseudoBraidToKrasnerGaussLib

if len(sys.argv) != 2:
	print("Expecting one argument.")
	sys.exit()
if sys.argv[1] == "-":
	x = sys.stdin.read()
else:
	x = sys.argv[1]
braid = [int(j) for j in re.findall("-?[0-9]+", x)]
print(str(pseudoBraidToKrasnerGaussLib.pseudoBraidToKrasnerGaussMain(braid)).replace(" ",""))



#!/usr/bin/python3
## @file

#
#    khoca/__init__.py --- Part of khoca, a knot homology calculator
#
# Copyright (C) 2021 Sebastian Oehms <seb.oehms@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# This file contains a Python wrapper class of the command line version
# of Khoca

# The main work is done by pui.pyx via run_commandline, which is called
# by the instance call of this class.

from khoca.khoca import run_commandline
from khoca.bin.pui import PyComplexStack


class InteractiveCalculator:
    r"""
    Class to allow the usage of `Khoca` interactively in a Python session.

    EXAMPLES::

        >>> from khoca import InteractiveCalculator
        >>> KH = InteractiveCalculator()
        >>> KH
        Khovanov homology calculator for Frobenius algebra: Z[X] / (1*X^2).
        >>> KH('braidaBaB')    # doctest: +NORMALIZE_WHITESPACE
        [[[-2, 4, 0, 1], [-1, 2, 0, 1], [0, 0, 0, 1], [1, -2, 0, 1], [2, -4, 0, 1],
          [-2, 4, 0, 0], [-1, 4, 0, 0], [-1, 2, 0, 0], [0, 2, 0, 0], [0, 0, 0, 0],
          [1, 0, 0, 0], [1, -2, 0, 0], [2, -2, 0, 0]], [[-2, 3, 0, 1], [-2, 5, 0, 1],
          [-1, 1, 0, 1], [-1, 3, 0, 1], [0, -1, 0, 1], [0, 1, 0, 1], [1, -3, 0, 1],
          [1, -1, 0, 1], [2, -5, 0, 1], [2, -3, 0, 1], [-1, 3, 2, 1], [-2, 3, 0, -1],
          [-1, 3, 0, -1], [-2, 5, 0, 0], [-1, 5, 0, 0], [-1, 1, 0, 0], [0, 1, 0, 0],
          [-1, 3, 0, 0], [0, 3, 0, 0], [0, -1, 0, 0], [1, -1, 0, 0], [0, 1, 0, 0],
          [1, 1, 0, 0], [2, -3, 2, 1], [1, -3, 0, -1], [2, -3, 0, -1], [1, -1, 0, 0],
          [2, -1, 0, 0]]]
        >>> res, mess = KH('braidaaa', print_messages=True); mess # doctest: +NORMALIZE_WHITESPACE
        ['Result:', 'Reduced Homology:', 'Non-equivariant homology:',
         't^-3q^8 + t^-2q^6 + t^0q^2', 'Unreduced Homology:',
         'Non-equivariant homology:', 't^-3q^9 + t^-2q^5 + t^0q^1 + t^0q^3 + t^-2q^7[2]']
        >>> KH((1, 1, 1)) == KH('braidaaa')
        True
        >>> KH([[4,2,5,1],[8,6,1,5],[6,3,7,4],[2,7,3,8]]) == KH('braidaBaB')
        True
        >>> KH((1, -2, 1, -2)) == KH('braidaBaB')
        True
        >>>
    """
    def __init__(self, coefficient_ring=0, frobenius_algebra=(0, 0), root=0, equivariant=None):
        r"""
        Constructor.

        INPUT:

            - ``coefficient_ring`` -- coefficient ring ``F``of the homology. It
              can be given by an integer (or string convertible to an integer):
              -  ``0`` (default) the ring of integers
              -  ``1``  the rational field
              -  a prime for the corresponding finite field
            - ``frobenius_algebra`` -- the Frobenius algebra ``F[x]/p`` given by
              the coefficients of the normed polynomial ``p`` (default is
              ``p = X^2``) as a tuple of length ``deg(p)`` where the constant
              term is the first entry.
            - ``root`` -- a root of the polynomial ``p`` (default is zero).
            - ``equivariant`` optional integer ``n > 1`` giving the degree of
              ``sl(n)`` for equivariant homology. If this is given then the input
              to the keyword arguments  ``frobenius_algebra`` and ``root`` will
              be ignored. If it is not given then non equivariant homology will
              be calculated.

        EXAMPLES::

            >>> from khoca import InteractiveCalculator
            >>> InteractiveCalculator(1, (0, 1), 0)
            Khovanov homology calculator for Frobenius algebra: Q[X] / (1*X^2 + 1*X).
            >>> from khoca import InteractiveCalculator
            >>> KH = InteractiveCalculator(equivariant=3); KH
            Khovanov homology calculator for Frobenius algebra: Z[a, b, c][X] / (1*X^3 + c*X^2 + b*X + a).
        """
        if type(coefficient_ring) == str:
            self._coefficient_ring = coefficient_ring
        else:
            try:
                self._coefficient_ring = str(int(coefficient_ring))
            except TypeError:
                raise TypeError('Coefficient ring must be declared by an integer or string')
        if type(frobenius_algebra) == str:
            self._frobenius_algebra = frobenius_algebra
        else:
            try:
                self._frobenius_algebra = '%s' %list(frobenius_algebra)
            except TypeError:
                raise TypeError('Frobenius algebra must be declared by a tuple, list or string')
        if type(root) == str:
            self._root = root
        else:
            try:
                self._root = str(int(root))
            except TypeError:
                raise TypeError('Root must be declared by an integer or string')

        self._equivariant = None
        if equivariant:
            try:
                 equivariant = int(equivariant)
            except TypeError:
                raise TypeError('equivariant must be given as integer')
            if equivariant <= 1:
                raise ValueError('equivariant must be larger than 1')
            self._equivariant = 'e%s' %equivariant

        self._description = None
        self._stack = PyComplexStack()
        self('braidaA', ' ') # sets self._description

    def __repr__(self):
        r"""
        String representation of ``self``.

        EXAMPLES::

            >>> from khoca import InteractiveCalculator
            >>> InteractiveCalculator(5, (1, 0), 0)
            Khovanov homology calculator for Frobenius algebra: F_5[X] / (1*X^2 + 1).
        """
        return 'Khovanov homology calculator for %s' %self._description


    def __call__(self, link, command=None, print_messages=False, verbose=False, progress=False):
        r"""
        Instance call to apply the calculator to a link with respect to a
        certain command.

        INPUT:

            - ``link`` -- the link to which the calculation should be done.
              It can be given as a Tuple which will be interpreted as a braid
              in Tietze form which closure is the link in question. Further
              a list of lists is accepted which will be interpreted as a list
              of crossings in pd-notation. Alternatively you can declare the
              link by a string. The following formats are accepted:

              - ``BraidX`` for a braid formatted as in knotscape (``a`` = first
                Artin generator, ``A`` = its inverse, ``b`` = second Artin
                generator, etc.). This works only for ``sl(2)`` homology,
                otherwise output is nonsensical
              - ``PdX`` for input in PD notation (as e.g. given on !KnotInfo).
                Again, this works only for ``sl(2)`` homology, otherwise output
                is nonsensical
              - ``GaussX`` for bipartite knot given as a matched diagram,
                following the convention explained in the section below. This
                works for ``sl(N)`` homology for all ``N``

            - ``command`` -- a command given as string as explained in the
              command line version of the functionality. This can be:

              - ``MirrorX`` takes the dual of the result at spot ``X``
              - ``SumXY`` computes the homology of the connected sum of the
                results saved at spots ``X`` and ``Y`` (numbers separated by
                a non-digit character)
              - ``CalcX` outputs the result saved at spot X. If you forget
                this command, the program will have no output

              If this keyword argument is not given it will be interpreted
              as ``Calc0`` by default.

              - ``print_messages`` boolean (default is ``False``). If set to
                ``True`` the print output of the command line version is
                returned as a list on the second output position. By default
                all print messages to ``stdout`` of the command line version
                are suppressed
              - ``verbose`` boolean (default is ``False``). If it is set to
                ``True`` all print messages to ``stdout`` together with special
                verbose messages of the command line version are printed
              - ``progress`` boolean (default is ``False``). If it is set to
                ``True`` progress bars will be printed

        OUTPUT:

            Two lists of quadruples the first for reduced and the second for
            unreduced homology. In such a quadruple the first item stands for
            the ``t``-degree, the second for the ``q``-degree, the third for
            the torsion and the last item stands for the coefficient of the
            corresponding summand of the homology.

            If the option ``print_messages`` is given, than the output contains
            a second result which is a list of all suppressed print messages.


        ..NOTE:

            The program keeps a stack of computed homologies, enumerated 0,1,2... .
            Each link declaration and each command except ``CalcX`` puts a new
            homology on that stack, whereas ``CalcX`` prints the homology at a
            certain spot. This is mainly useful to compute homology of sums of
            knots. As an example of such a stack operations see the first example
            below.

        EXAMPLES::

            >>> from khoca import InteractiveCalculator
            >>> KH = InteractiveCalculator(1, (1, 0), 0); KH
            Khovanov homology calculator for Frobenius algebra: Q[X] / (1*X^2 + 1).
            >>> KH('braidaaa', 'dual0 sum0+1 braidaBaB sum2+3 calc4', print_messages=True) # doctest: +NORMALIZE_WHITESPACE
            ([[[-5, 10, 0, 1], [-4, 8, 0, 1], [-4, 8, 0, 1], [-3, 6, 0, 1],
               [-3, 6, 0, 1], [-3, 6, 0, 1], [-2, 4, 0, 1], [-2, 4, 0, 1],
               [-2, 4, 0, 1], [-2, 4, 0, 1], [-2, 4, 0, 1], [-2, 4, 0, 1],
               [-1, 2, 0, 1], [-1, 2, 0, 1], [-1, 2, 0, 1], [-1, 2, 0, 1],
               [-1, 2, 0, 1], [-1, 2, 0, 1], [-1, 2, 0, 1], [0, 0, 0, 1],
               [0, 0, 0, 1], [0, 0, 0, 1], [0, 0, 0, 1], [0, 0, 0, 1],
               [0, 0, 0, 1], [0, 0, 0, 1], [1, -2, 0, 1], [1, -2, 0, 1],
               [1, -2, 0, 1], [1, -2, 0, 1], [1, -2, 0, 1], [1, -2, 0, 1],
               [1, -2, 0, 1], [2, -4, 0, 1], [2, -4, 0, 1], [2, -4, 0, 1],
               [2, -4, 0, 1], [2, -4, 0, 1], [2, -4, 0, 1], [3, -6, 0, 1],
               [3, -6, 0, 1], [3, -6, 0, 1], [4, -8, 0, 1], [4, -8, 0, 1],
               [5, -10, 0, 1]], [[0, -1, 0, 1], [0, 1, 0, 1]]],
             ['Result:', 'Reduced Homology:', 'Non-equivariant homology:',
              'Page 1:', 't^-5q^10 + 2t^-4q^8 + 3t^-3q^6 + 6t^-2q^4 + 7t^-1q^2 + 7t^0q^0
                       + 7t^1q^-2 + 6t^2q^-4 + 3t^3q^-6 + 2t^4q^-8 + t^5q^-10',
              'The spectral sequence collapses on the first page.\n',
              'Unreduced Homology:', 'Non-equivariant homology:',
              'Page 1:', 't^-5q^11 + t^-4q^7 + t^-4q^9 + t^-3q^5 + 2t^-3q^7 + 2t^-2q^3
                       + 4t^-2q^5 + 4t^-1q^1 + 3t^-1q^3 + 4t^0q^-1 + 4t^0q^1 + 3t^1q^-3
                       + 4t^1q^-1 + 4t^2q^-5 + 2t^2q^-3 + 2t^3q^-7 + t^3q^-5 + t^4q^-9
                       + t^4q^-7 + t^5q^-11',
              'Page 3 = infinity:', 't^0q^-1 + t^0q^1'])
        """
        # Todo: allow Python type inputs like tuples ..
        from .khoca import set_options
        verbose = int(verbose); progress = int(progress)
        set_options(verbose, progress)
        arg_list = [None]
        arg_list.append(self._coefficient_ring)
        if self._equivariant:
            arg_list.append(self._equivariant)
        else:
            arg_list.append(self._frobenius_algebra)
        arg_list.append(self._root)


        if type(link) in (tuple, list):
            # Allow more Python like ways to declare the link
            try:
                link_list = [int(j) for j in link]
                # interpreted as braid in Tietze-form
                offset = {-1:64, 1:96}
                from math import copysign
                link_knotscape = ''
                for j in link_list:
                    link_knotscape += chr(abs(j) + offset[copysign(1,j)])
                link = 'braid' + link_knotscape
            except TypeError:
                try:
                    link_list = [list(j) for j in link]
                    # interpreted as pd-Code
                    link = 'pd' + str(link_list)
                except TypeError:
                    pass

        arg_list.append(link)

        if not command:
            command = 'calc0'
        cmd_list = command.split()
        arg_list += cmd_list

        # redirect print outputs to a list to
        # return it optionally to the user
        print_output = []
        def no_print(s):
            print_output.append(s)

        if verbose:
            if print_messages:
                raise ValueError('All messages are printed in verbose mode')
            with self._stack:
                res = run_commandline(arg_list, print, progress)
            return res

        # capture print output from the shared library
        # and use the to set the description resp.
        # return it optionally to the user
        import py
        capture = py.io.StdCaptureFD(err=False, in_=False)

        try:
            with self._stack:
                res = run_commandline(arg_list, no_print, progress)
        except:
            out, err = capture.reset()
            raise

        out, err = capture.reset()
        if not self._description and out:
            self._description = out.strip('\n')

        if print_messages:
            if  out:
                print_output += [out][:-1]
            return res, print_output
        return res

    def khoca_version(self):
        r"""
        Return the version if ``Khoca``.

        EXAMPLES::

            >>> from khoca import InteractiveCalculator
            >>> KH = InteractiveCalculator()
            >>> KH.khoca_version() >= '1.1'
            True
        """
        from khoca._version import version
        return version



#!/usr/bin/python3

#
#    pseudoBraidToKrasnerGaussLib.py --- Part of khoca, a knot homology calculator
#
# Copyright (C) 2018 Lukas Lewark <lukas@lewark.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


import sys

def pseudoBraidToKrasnerGaussMain(inputList):
	strands = max(inputList) + 1
	if (strands % 2 == 1):
		strands += 1

	direction = -1
	enum = [0 for i in range(len(inputList))]
	result = [0 for i in range(2*len(inputList))]
	num = 1
	for i in range(strands):
		idx = [j for j in range(len(inputList)) if (abs(inputList[j]) in [i, i+1])]
		if (direction == -1):
			idx.reverse()
		direction *= -1
		for j in idx:
			if (abs(inputList[j]) == i+1):
				enum[j] = num
			else:
				result[enum[j] - 1] = num * inputList[j] / abs(inputList[j])
			num += 1

	return [int(i) for i in result if (i != 0)]



#!/usr/bin/python3
## @file

#
#    khoca.py --- Part of khoca, a knot homology calculator
#
# Copyright (C) 2018 Lukas Lewark <lukas@lewark.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# The file to be run by the user.
# The main work is done by pui.pyx, which is called by this file.

import sys
sys.path.insert(0, './bin/')

import os

def interactive():
    r"""
    Checks if the module is used as part of :class:`InteractiveCalculator`.
    """
    head, tail = os.path.split(__file__)
    if not head:
        return False
    head, tail = os.path.split(head)
    if tail == 'khoca':
        return True
    else:
        return False

def run_command_exit(print_command, err_mess):
    r"""
    Exit :func:`run_command_line` on error.
    """
    if interactive():
        raise ValueError(err_mess)
    else:
        print_command(err_mess)
        sys.exit()

def set_options(verbose_mode, progress_mode):
    r"""
    Set the global options ``verbose`` and ``progress``.
    """
    global verbose, progress
    verbose = verbose_mode
    progress = progress_mode

import random
import re
import math

sys.path.append('.')
if interactive():
    from khoca.bin.pui import pCalcSubTangleTree, pCalculateHomology, pGlueReduced, pCopyHomology, pFirstFreeIdx, pSaveComplexToFile, pDeleteComplex, pLoadComplexFromFile, pSum, pDual, pDeleteNonIsos, pResetSimplificationsCounter, pReducify, pIniStack, pSetRoot, pPrintHomology, pPrintCompileInfo, pCalculateEquiHomology
    from khoca.bin import KrasnerGaussToMyPDLib, pseudoBraidToKrasnerGaussLib, BraidToMyPD
else:
    from pui import pCalcSubTangleTree, pCalculateHomology, pGlueReduced, pCopyHomology, pFirstFreeIdx, pSaveComplexToFile, pDeleteComplex, pLoadComplexFromFile, pSum, pDual, pDeleteNonIsos, pResetSimplificationsCounter, pReducify, pIniStack, pSetRoot, pPrintHomology, pPrintCompileInfo, pCalculateEquiHomology
    import KrasnerGaussToMyPDLib, pseudoBraidToKrasnerGaussLib, BraidToMyPD


debugging = False
NUM_THREADS = 1 if debugging else 12


## Converts a link diagram given in the Planar-diagram notation to the mypd format.
# @param pd A list of lists of four integers each, such as [[0,1,2,3],[1,4,5,2],[4,0,3,5] for the positive trefoil.
# @return That link diagram in mypd-notation, type "0" for positive, type "1" for negative crossings.
def pd_to_mypd(pd):
    result = list()
    for i in pd:
        if (i[3] == ((i[1] + 1) % (2 * len(pd)))):
            result.append([2,i[0] - 1,i[1] - 1,i[2] - 1,i[3] - 1])
        else:
            result.append([3,i[1] - 1,i[2] - 1,i[3] - 1,i[0] - 1])
    return result

def generate_binary_tree(length, s = 0):
    if (length == 1):
        return s
    return [generate_binary_tree(length / 2, s), generate_binary_tree((length / 2) + (length % 2), s + (length / 2))]

def generate_standard_tree(l):
    if (len(l) == 1):
        return l[0]
    return [generate_standard_tree(l[:-1]), l[len(l) - 1]]

def isBetterSignature(s1, s2):
    s1Copy = list(s1)
    s2Copy = list(s2)
    s1Copy.sort(reverse = True)
    s2Copy.sort(reverse = True)
    if (s1Copy < s2Copy): return True
    if (s1Copy > s2Copy): return False
    return (s1 > s2)

# Using a randomised algorithm, produces a linear tree such that intermediate tangles have small girth
def generate_girth_optimised_tree(pd):
    pd = [[i] + pd[i][1:] for i in range(len(pd))]
    iterations = 1000

    bestOrder = []
    bestSignature = []
    girth = 0
    for k in range(iterations):
        copyPd = list(pd)
        openEnds = set()
        order = []
        signature = []
        while len(copyPd) > 0:
            girthChange = [4 - 2 * len(openEnds & set(i[1:])) for i in copyPd]
            bestGirthChange = min(girthChange)
            bestCrossings = [i for i in range(len(girthChange)) if girthChange[i] == bestGirthChange]
            newCrossing = bestCrossings[0] if debugging else random.choice(bestCrossings)
            order.append(copyPd[newCrossing][0])
            openEnds = openEnds ^ set(copyPd.pop(newCrossing)[1:])
            signature.append(len(openEnds))
        if (bestOrder == []) or (isBetterSignature(signature, bestSignature)):
            bestOrder = order
            bestSignature = signature
            girth = max(bestSignature)
    return (girth, generate_standard_tree(bestOrder))

def eliminatedoubles(l):
    i = 0
    while (i < len(l)):
        j = i + 1
        found = False
        while (j < len(l)):
            if l[i] == l[j]:
                l.pop(j)
                l.pop(i)
                found = True
                break
            j = j + 1
        if not found:
            i = i + 1

## Stores the information which elementary tangles to load and how and in which order to glue them.
# In particular, all index shifts have been done, so that calcSubTangleTree may
# call functions in pythonInterface.cpp without further ado.
# Details: This represents a directed tree with vertices numbered 0,1,....
# Each vertex of number i has at most two children, whose number is saved in
# sons[i] and daughers[i] (-1 meaning no child). types saves whether the tangle at the vertex
# is elementary and should be loaded from a file (then types[i] is 0,1,2,3, see function loadType in pui.pyx),
# or the tangle is obtained by gluing son and daughter (then types[i] is -1).
# boundaryPoints[i] is an orderd list of the tangles' boundary points (in global numbering).
# when gluing e.g. a daughter and son with boundaryPoints 2,3,0,1 and 4,6,3,2, respectively, one
# first takes disjoint union (giving a tangle with boundaryPoints 2,3,0,1,4,6,3,2), and then
# glues the matching boundaryPoints; which match which is saved in daughterGluePoints and sonGluePoints,
# which are 0,0 and 7,5 here (because index shifts are already contained).
class CSchedule:
    def initialise(self, mypd):
        nodes = 2 * len(mypd) - 1
        self.shift = 0
        self.daughters = [-1] * nodes
        self.sons = [-1] * nodes
        self.types = [-1] * nodes
        self.selfGlue = [list() for i in range(nodes)]
        self.boundaryPoints = [list() for i in range(nodes)]
        self.sonGluePoints = [list() for i in range(nodes)]
        self.daughterGluePoints = [list() for i in range(nodes)]
        return nodes + 1

    def output(self):
        sys.stderr.write(str(self.daughters) + str(self.sons) + str(self.types) + str(self.boundaryPoints) + str(self.daughterGluePoints) + str(self.sonGluePoints) + str(self.selfGlue) + "\n")


    def getFromMypdAndTree(self, mypd, tree, idx):
        if (type(tree) == type(int())):
            self.types[idx] = mypd[tree][0]
            self.boundaryPoints[idx] = mypd[tree][1:]
            unitedBoundaryPoints = list(self.boundaryPoints[idx])
            i = 0
            while (i < len(unitedBoundaryPoints)):
                j = i + 1
                found = False
                while (j < len(unitedBoundaryPoints)):
                    if unitedBoundaryPoints[i] == unitedBoundaryPoints[j]:
                        self.selfGlue[idx].append([i,j])
                        unitedBoundaryPoints.pop(j)
                        unitedBoundaryPoints.pop(i)
                        found = True
                        break
                    j = j + 1
                if not found:
                    i = i + 1
            return idx + 1
        else:
            self.daughters[idx] = idx + 1
            self.sons[idx] = self.getFromMypdAndTree(mypd, tree[0], idx + 1)
            result = self.getFromMypdAndTree(mypd, tree[1], self.sons[idx])
            beforeDeletion = (self.boundaryPoints[self.daughters[idx]] + self.boundaryPoints[self.sons[idx]])
            self.boundaryPoints[idx] = [ x for x in beforeDeletion if beforeDeletion.count(x) == 1]
            shift = len(self.boundaryPoints[self.daughters[idx]])
            daughterBP = list(self.boundaryPoints[self.daughters[idx]])
            eliminatedoubles(daughterBP)
            sonBP = list(self.boundaryPoints[self.sons[idx]])
            eliminatedoubles(sonBP)
            unitedBoundaryPoints = list(daughterBP)
            unitedBoundaryPoints.extend(sonBP)
            i = 0
            while (i < len(unitedBoundaryPoints)):
                j = i + 1
                found = False
                while (j < len(unitedBoundaryPoints)):
                    if unitedBoundaryPoints[i] == unitedBoundaryPoints[j]:
                        self.daughterGluePoints[idx].append(i)
                        self.sonGluePoints[idx].append(j)
                        unitedBoundaryPoints.pop(j)
                        unitedBoundaryPoints.pop(i)
                        found = True
                        break
                    j = j + 1
                if not found:
                    i = i + 1
            return result

def reducePD(mypd):
    x = [item for subl in mypd for item in subl[1:]]
    j = x.index(max(x))
    mypd[int(j / 4)][1 + j % 4] +=1
    return mypd

def calcIt(mypd, shift):
    global verbose, progress
    (girth, tree) = generate_girth_optimised_tree(mypd)
    pIniStack(stackMod, stackFrobenius, N, girth, verbose)
    pSetRoot(stackRoot)

    s = CSchedule()
    result = s.initialise(mypd)
    s.getFromMypdAndTree(mypd, tree, 0)
    s.shift = shift

    pCalcSubTangleTree(s, NUM_THREADS, progress)
    return result

def getInts(s, signed):
    return [int(j) for j in re.findall( ("-?" if signed else "") + "[0-9]+", s[3:])]

## @var mypd
# The internal format for a closed object (as a link), glued together from smaller tangles (as tangles).
#
# mypd is a list of tangles; an object is a list of integers.
# The first of that integers is the type of the object (e.g. a positive or a
# negative crossing), all following integers are the numbers of the boundary
# points of that object, the order of which matters.
# %Boundary point numbers need not be consecutive; each boundary point number
# must appear exactly twice in the whole mypd.

def isAllowedStackMod(a):
    if not a.isdigit():
        return False
    x = int(a)
    if (x <= 3):
        return True
    for i in range(2, int(math.sqrt(x) + 2)):
        if (x % i == 0):
            return False
    return True

def run_commandline(argv, printCommand, progress):
        NUM_ARGUMENTS = 3
        global stackMod, stackFrobenius, N, stackRoot

        # Parsing the arguments
        if (len(argv) <= NUM_ARGUMENTS):
            printCommand(HELP_TEXT)
            return 1

        if isAllowedStackMod(argv[1]):
            stackMod = int(argv[1])
        else:
            printCommand("First argument must be 0, 1 or a prime.")
            return 1

        N = 0
        if argv[2][0] == "e":
            N = int(argv[2][1:])
            l = []
            if (N < 2):
                printCommand("e must be followed by an integer that is at least 2.")
                return 1
        else:
            l = [int(i) for i in re.findall("-?[0-9]+", argv[2])]
            if (len(l) == 0):
                printCommand("Second argument must be a list of at least one integer, or e (for equivariant).")
                return 1
        stackFrobenius = l
        equivariant = (len(stackFrobenius) == 0)

        if argv[3].isdigit() or ((argv[3][0] in ["+","-"]) and (argv[3][1:].isdigit())):
            stackRoot = int(argv[3])
        else:
            printCommand("Third argument must be a signed integer.")
            return 1

        # doing the work
        shift = 0
        idxTranslator = [0]
        res = []
        for i in argv[(NUM_ARGUMENTS + 1):]:
            if i[:3].capitalize() == "Nat":
                l = getInts(i, False)
                if (len(l) % 5 != 0) or (len(l) == 0):
                    run_command_exit(printCommand, "The native code (\"" + i[3:] +  "\") following the command \"nat\" must be a non-empty list of integers whose length is divisible by five.")

                mypd = [list(j) for j in zip(*[iter(l)]*5)]
                mypd = reducePD(mypd)
                shift += calcIt(mypd, shift)
                idxTranslator += [shift]
            if i[:7].capitalize() == "Special":
                l = eval(i[7:])
                shift += calcIt(l, shift)
                idxTranslator += [shift]
            elif i[:5].capitalize() == "Gauss":
                l = getInts(i, True)
                mypd = reducePD(KrasnerGaussToMyPDLib.KrasnerGaussToMyPDmain(l))
                shift += calcIt(mypd, shift)
                idxTranslator += [shift]
            elif i[:11].capitalize() == "PseudoBraid":
                m = getInts(i, True)
                l = pseudoBraidToKrasnerGaussLib.pseudoBraidToKrasnerGaussMain(m)
                mypd = reducePD(KrasnerGaussToMyPDLib.KrasnerGaussToMyPDmain(l))
                shift += calcIt(mypd, shift)
                idxTranslator += [shift]
            elif i[:5].capitalize() == "Braid":
                mypd = reducePD(BraidToMyPD.BraidToMyPD(i[5:]))
                shift += calcIt(mypd, shift)
                idxTranslator += [shift]
            elif i[:2].capitalize() == "Pd":
                l = getInts(i, False)
                l = [x - 1 for x in l]
                mypd = reducePD(pd_to_mypd([list(j) for j in zip(*[iter(l)]*4)]))
                print(mypd)
                shift += calcIt(mypd, shift)
                idxTranslator += [shift]
            elif i[:10].capitalize() == "Calcnonred":
                l = getInts(i, False)
                if (len(l) != 1):
                    run_command_exit(printCommand, "\"Calcnonred\" must be followed by exactly one unsigned integer.")
                idx = idxTranslator[l[0]]
                firstFreeIdx = shift
                shift += 1
                idxTranslator[-1] += 1
                pResetSimplificationsCounter(idx)
                pCopyHomology(idx, firstFreeIdx)
                printCommand("Result:")
                printCommand("Unreduced Homology:")
                pGlueReduced(firstFreeIdx, 1, NUM_THREADS, progress)
                s = pCalculateHomology(firstFreeIdx, False, NUM_THREADS, printCommand, equivariant, progress)
                res.append(s)
                pDeleteComplex(firstFreeIdx)
            elif i[:4].capitalize() == "Calc":
                nice = (i[4:8].capitalize() == "Nice")
                l = getInts(i, False)
                if (len(l) != 1):
                    run_command_exit(printCommand, "\"Calc\" must be followed by exactly one unsigned integer.")
                idx = idxTranslator[l[0]]
                firstFreeIdx = shift
                secondFreeIdx = shift + 1
                shift += 2
                idxTranslator[-1] += 2
                pResetSimplificationsCounter(idx)
                pCopyHomology(idx, firstFreeIdx)
                pCopyHomology(idx, secondFreeIdx)
                printCommand("Result:")
                printCommand("Reduced Homology:")
                pReducify(secondFreeIdx)
                s = pCalculateHomology(secondFreeIdx, nice, NUM_THREADS, printCommand, equivariant, progress, True)
                res.append(s)
                printCommand("Unreduced Homology:")
                pGlueReduced(firstFreeIdx, 1, NUM_THREADS, progress)
                s = pCalculateHomology(firstFreeIdx, nice, NUM_THREADS, printCommand, equivariant, progress, True)
                res.append(s)
                pDeleteComplex(firstFreeIdx)
                pDeleteComplex(secondFreeIdx)
            elif i[:4].capitalize() == "Save":
                l = re.match("[0-9]+", i[4:])
                if (l == None):
                    run_command_exit(printCommand, "\"Save\" must be followed by an unsigned integer and a filename.")
                pSaveComplexToFile(idxTranslator[int(l.group(0))], i[(4 + len(l.group(0))):])
            elif i[:4].capitalize() == "Load":
                pLoadComplexFromFile(shift, i[4:])
                shift += 1
                idxTranslator += [shift]
            elif i[:4].capitalize() == "Dual":
                l = re.match("[0-9]+", i[4:])
                if (l == None):
                    run_command_exit(printCommand, "\"Dual\" must be followed by an unsigned integer.")
                pDual(shift, int(l.group(0)))
                shift += 1
                idxTranslator += [shift]
            elif i[:3].capitalize() == "Sum":
                l = getInts(i, False)
                if (len(l) != 2):
                    run_command_exit(printCommand, "\"Sum\" must be followed by exactly two unsigned integer.")
                pSum(shift, idxTranslator[l[0]], idxTranslator[l[1]], NUM_THREADS, progress)
                shift += 1
                idxTranslator += [shift]
            else:
                run_command_exit(printCommand, str(i) + " is not a valid command.")
        return res


def parseOptions(argv):
    global verbose, progress
    o = [x for x in argv[1:] if (x[0] == '-')]
    for i in o:
        if (i == '-v'):
            verbose = 1
        elif (i == '-p'):
            progress = 1
        elif (i == '-h'):
            print(HELP_TEXT)
            sys.exit()
        else:
            print("Unknown option " + i + ".")
            sys.exit()
    return [argv[0]] + [x for x in argv[1:] if (x[0] != '-')]

HELP_TEXT = "Expecting at least three arguments:\n(1) The coefficient ring; 0 for integers, 1 for rationals, a prime p for the finite field with p elements,\n(2) the vector [a_0, ... a_{N-1}] defining the Frobenius algebra F[X]/(X^N+a_{N-1}X^{N-1}+...+a_0) or e followed by the number N for equivariant,\n(3) a root of the polynomial given in (2).\nAny following argument is interpreted as command. For example,\n./khoca.py 0 0.0 0 braidaBaB calc0\ncomputes integral Khovanov sl(2)-homology of the figure-eight knot.\nRefer to the README file or to http://lewark.de/lukas/khoca.html for more detailed help."
verbose = 0
progress = 0

if (verbose):
    pPrintCompileInfo()
    sys.stderr.write(("Multithreading with " + str(NUM_THREADS) + " threads.\n") if (NUM_THREADS > 1) else "No multithreading.\n")

if not interactive():
    a = parseOptions(sys.argv)
    run_commandline(a, print, progress)



#!/usr/bin/python3

#
#    montesinos.py --- Part of khoca, a knot homology calculator
#
# Copyright (C) 2018 Lukas Lewark <lukas@lewark.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


from math import gcd
import sys, re

def sgn(x):
	assert(x != 0)
	return 1 if x > 0 else -1

# Increase last entry of a list x by d, return the result
def changeLast(x, d):
	assert(len(x) > 0)
	return [i for i in x[:-1]] + [x[len(x)-1] + d]

# Change sign of all entries of a list, return result
def neg(x):
	return [-i for i in x]

# Accepts as input two integers p, q, with q non-zero.
# Returns a continued fraction [a_1, ... a_n] = a_n + 1/(a_(n-1) + ...)
# for the rational p / q, where a_1, ... a_(n-1) are even numbers,
# and a_n is even if p or q is, and odd if p and q are.
def cf(p,q):
	if (p == 0): return [0]
	if (p < 0): return neg(cf(-p,q))
	if (p > q): return changeLast( cf( ((p + q) % (2*q)) - q, q), 2 * int((p + q) / (2*q)))
	if (((p * q) % 2) == 1): return changeLast( neg(cf(q - p, q)), 1)
	return changeLast(neg(cf(2*p - q, p)), 2) + [0]

def parse(s):                                 
	reg = re.compile("(-?[0-9]+)(/[0-9]+)?")
	x = reg.findall(sys.argv[1])
	return [[int(i[0]), 1 if (i[1] == "") else int(i[1][1:])] for i in x]

# Read from input
l = parse(sys.argv[1])
evenDen = None
j = 0
for i in l:
	assert(i[1] != 0)
	g = gcd(i[0],i[1])
	i[0] = int(i[0] / g)
	i[1] = int(i[1] / g)
	if (j == 0) and (len(l) == 1) and (l[0][0] % 2 == 1) and (l[0][1] % 2 == 1):
		l[0][1] += l[0][0]
	if (i[1] % 2 == 0) and (evenDen == None):
		evenDen = j
	j += 1
del j

L = [cf(l[i][0], l[i][1]) for i in range(len(l)) if i != evenDen]
s = sum([i[len(i) - 1] for i in L])
for i in L:
	i[len(i) - 1] = 0
if (evenDen != None):
	L.insert(evenDen, cf(l[evenDen][0] + l[evenDen][1] * s, l[evenDen][1]))
	s = 0
del l, evenDen
evens = [i for i in range(len(L)) if (len(L[i]) % 2 == 0)]
if not (((len(evens) == 1) and (s % 2 == 0)) or ((len(evens) == 0) and (s % 2 == 1))):
	print("This is not a knot.")
	sys.exit()
if (s % 2 != 0):
	print("Cannot produce a matched diagram.")
	sys.exit()
# cycle even to the front
L = L[evens[0]:] + L[:evens[0]]
# adapt sign (now: handedness, after: sign of crossing)
for i in L:
	for j in range(len(i)):
		if ((j % 2) == (len(i) % 2)):
			i[j] *= -1
                       
# part coming from s

T = [[[] for j in range(int(sum([abs(k) for k in i]) / 2))] for i in L]
T.insert(0, [[] for i in range(int(abs(s) / 2))])
    
#part coming from the even tangle

shift = 0
for i in range(int(abs(s) / 2)):
	T[0][i].append(sgn(s) * (shift + i + 1))
shift += int(abs(s) / 2)

minishift = 0
for i in range(int(len(L[0]) / 2)):
	for j in range(int(abs(L[0][2 * i]) / 2)):
		T[1][minishift + j].append(sgn(L[0][2 * i])*(j + shift + 1))
	shift += int(abs(L[0][2*i]) / 2)
	minishift += int(abs(L[0][2 * i + 1]) / 2) + int(abs(L[0][2*i]) / 2)

for i in range(int(abs(s) / 2)):
	T[0][int(abs(s) / 2) - i - 1].append(sgn(s) * (shift + i + 1))
shift += int(abs(s) / 2)

for i in range(len(L)):
	minishift = 0
	for j in range(len(L[len(L) - i - 1])):
		for k in range(int(abs(L[len(L) - i - 1][-j-1]) / 2)):
			T[len(L) - i][-k - minishift-1].append(sgn(L[len(L) - i - 1][-1-j])*(k + shift + 1))
		minishift += int(abs(L[len(L) - i - 1][-1-j]) / 2)
		shift += int(abs(L[len(L) - i - 1][-1-j]) / 2)
	minishift = 0
	for j in range(int(len(L[len(L) - i - 1]) / 2)):
		minishift += int(abs(L[len(L) - i - 1][2 * j]) / 2)
		for k in range(int(abs(L[len(L) - i - 1][2 * j + 1]) / 2)):
			T[len(L) - i][k + minishift].append(sgn(L[len(L) - i - 1][2 * j + 1])*(k + shift + 1))
		minishift += int(abs(L[len(L) - i - 1][2 * j + 1]) / 2)
		shift += int(abs(L[len(L) - i - 1][2 * j + 1]) / 2)

for i in range(len(L) - 1):
	minishift = 0
	for j in range(int(len(L[i + 1]) / 2)):
		for k in range(int(abs(L[i + 1][2 * j]) / 2)):
			T[i + 2][k + minishift].append(sgn(L[i + 1][2 * j])*(k + shift + 1))
		minishift += int(abs(L[i + 1][2 * j + 1]) / 2) + int(abs(L[i + 1][2*j]) / 2)
		shift += int(abs(L[i + 1][2 * j]) / 2)

T2 = [[min(abs(i[0]),abs(i[1])),max(i) if (i[0] > 0) else min(i)] for j in T for i in j]
T2.sort()
T3 = [i[1] for i in T2]
print(str(T3).replace(" ",""))



#!/usr/bin/python3

#
#    humanToBin.py --- Part of khoca, a knot homology calculator
#
# Copyright (C) 2018 Lukas Lewark <lukas@lewark.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#



import sys, re, math

qShift_t = -16
tShift_t = qShift_t
dot_t = 16
vidx_t = 32
vidx2_t = 64
boundary_t = -8
uint32_t = 32

def myWrite(val, size):
	assert(size % 8 == 0)
	sys.stdout.buffer.write((val).to_bytes(int(abs(size) / 8), byteorder='little', signed=(size < 0)))

def myPop(size):
	res = nlist.pop(0)
	myWrite(res, size)
	return res

def mpirUInt(x):
	assert(x >= 0)
	if (x == 0):
		myWrite(0, 8)
	else:
		size = int(math.log(x, 256)) + 1
		myWrite(size, 8)
		myWrite(x, size * 8)
	

def mpirRat(coefficientRing):
	if (coefficientRing == 0):
		x = nlist.pop(0)
		if (x >= 0):
			myWrite(0, 8)
		else:
			myWrite(1, 8)
		myWrite(abs(x), 32)
	elif (coefficientRing == 1):
		num = nlist.pop(0)
		den = nlist.pop(0)
		if (num < 0):
			myWrite(1, 8)
		else:
			myWrite(0, 8)
		mpirUInt(abs(num))
		mpirUInt(den)
	elif (coefficientRing < 256):
		x = nlist.pop(0)
		myWrite(x, 8)
	else:
		x = nlist.pop(0)
		myWrite(x, 16)


def krasnerTangle(dataSize):
	myPop(qShift_t) # get qShift
	myPop(qShift_t) # get qShift times N
	myPop(boundary_t) # get Circle coun
	for i in range(dataSize):
		myPop(boundary_t)


def vecTangles(boundarySize):
	tanglesSize = myPop(vidx2_t)
	for i in range(tanglesSize):
		krasnerTangle(boundarySize)
	deloopStackSize = myPop(vidx2_t)
	for i in range(deloopStackSize):
		myPop(vidx2_t)

def krasnerCobo(coefficientRing):
	mpirRat(coefficientRing)
	
	size = myPop(boundary_t)
	for i in range(size):
		myPop(dot_t)

def LCCobos(coefficientRing):
	size = myPop(vidx2_t)
	for i in range(size):
		krasnerCobo(coefficientRing)

def matLCCobos(coefficientRing):
	colCount = myPop(uint32_t)
	entryCount = myPop(uint32_t)
	for i in range(entryCount):
		LCCobos(coefficientRing)
	for i in range(entryCount):
		myPop(uint32_t)
	rowPtrSize = myPop(uint32_t)
	for i in range(rowPtrSize):
		myPop(uint32_t)
	invertiblesSize = myPop(uint32_t)
	for i in range(invertiblesSize):
		myPop(uint32_t)

def complex():
	coefficientRing = myPop(tShift_t) # global T-shift
	myPop(qShift_t) # global T-shift
	boundarySize = myPop(boundary_t) # boundary size
	complexSize = myPop(vidx_t) # complex size ( = number of modules)
	for i in range(complexSize):
		vecTangles(boundarySize)
	for i in range(complexSize - 1):
		matLCCobos(coefficientRing)

nlist = [int(i) for i in re.findall(r'-?\d+', sys.stdin.read())];
complex()



#!/usr/bin/python3

#
#    KrasnerGaussToMyPDLib.py --- Part of khoca, a knot homology calculator
#
# Copyright (C) 2018 Lukas Lewark <lukas@lewark.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys, re

def inSublist(x, l):
	for i in range(len(l)):
		if x in l[i]:
			return i
	raise

def splitSublist(x, y, l):
	i = inSublist(x, l)
	idx1 = l[i].index(x)
	idx2 = l[i].index(y)
	minIdx = min(idx1, idx2)
	maxIdx = max(idx1, idx2)
	new1 = l[i][:minIdx] + l[i][maxIdx + 1:]
	new2 = l[i][minIdx + 1:maxIdx]
	l.pop(i)
	l.append(new1)
	l.append(new2)
	
def KrasnerGaussToMyPDmain(gauss):
	tmp = list(range(1, 2 * len(gauss) + 1))
	decompressed_gauss = [0] * 2 * len(gauss)
	signs = []
	for i in range(2 * len(gauss)):
		signs.append([True])
	mypd = []
	for i in gauss:
		x = tmp[0]
		decompressed_gauss[x - 1] = abs(i) - 1
		decompressed_gauss[abs(i) - 1] = x - 1
		if i < 0:
			signs[abs(i) - 1] = False
			signs[x - 1] = False
		tmp = tmp[1:]
		tmp.remove(abs(i))
	del tmp
	
	isInside = []
	for i in range(2 * len(gauss)):
		isInside.append([True, False])
	insidePart = [list(range(2 * len(gauss)))]
	outsidePart = [list(range(2 * len(gauss)))]
	idle = False
	
	while not idle:
		idle = True
		for i in range(2 * len(gauss)):
			if (len(isInside[i]) == 2):
				if inSublist(i, insidePart) != inSublist(decompressed_gauss[i], insidePart):
					isInside[i].remove(True)
					isInside[decompressed_gauss[i]].remove(True)
				if inSublist(i, outsidePart) != inSublist(decompressed_gauss[i], outsidePart):
					isInside[i].remove(False)
					isInside[decompressed_gauss[i]].remove(False)
				if isInside[i] == []:
					raise
				elif isInside[i] == [True]:
					splitSublist(i, decompressed_gauss[i], insidePart)
					idle = False
				elif isInside[i] == [False]:
					splitSublist(i, decompressed_gauss[i], outsidePart)
					idle = False
		if idle:
			for i in range(2 * len(gauss)):
				if (len(isInside[i]) == 2):
					isInside[i].remove(False)
					isInside[decompressed_gauss[i]].remove(False)
					splitSublist(i, decompressed_gauss[i], insidePart)
					idle = False
					break
	
	pd = []
	
	for i in range(len(decompressed_gauss)):
		x = decompressed_gauss[i]
		if i < x:
			toadd = [x, (x + 1) % (2 * len(gauss)), abs(i), (abs(i) + 1) % (2 * len(gauss))]
			toaddPd1 = [2 * x, 2 * abs(i) + 1, 2 * x + 1, (2 * abs(i) + 2) % (4 * len(gauss))]
			toaddPd2 = [2 * abs(i), 2 * x + 1, 2 * abs(i) + 1, (2 * x + 2) % (4 * len(gauss))]
	
			if isInside[i][0] and signs[i]:
				toaddPd1 = [toaddPd1[0], toaddPd1[3], toaddPd1[2], toaddPd1[1]]
				toaddPd2 = [toaddPd2[0], toaddPd2[3], toaddPd2[2], toaddPd2[1]]
			elif isInside[i][0] or signs[i]:
				toaddPd1 = [toaddPd1[1], toaddPd1[0], toaddPd1[3], toaddPd1[2]]
				toaddPd2 = [toaddPd2[1], toaddPd2[0], toaddPd2[3], toaddPd2[2]]
	
			if isInside[i][0]:
				toadd.reverse()
			if signs[i]:
				toadd = toadd[1:] + toadd[:1]
				toadd = [0] + toadd
			else:
				toadd = toadd[1:] + toadd[:1]
				toadd = [1] + toadd
			mypd.append(toadd)
			pd.append(toaddPd1)
			pd.append(toaddPd2)
	
	return mypd



#
#    braidToMyPD.py --- Part of khoca, a knot homology calculator
#
# Copyright (C) 2018 Lukas Lewark <lukas@lewark.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

def chrToInt(i):
	t = ord(i)
	if (t in range(97, 123)):
		return t - 96
	elif (t in range(65, 91)):
		return 64 - t
	else:
		print("Unallowed char in braid: \"" + i + "\"")
		return 0

def BraidToMyPD(s):
	l = [chrToInt(i) for i in s]
	assert(not 0 in l)
	stumps = list(range(max(max(l), -min(l)) + 1))
	m = len(stumps)
	r = []
	for i in l:
		c = [3, m, m + 1, stumps[abs(i) - 1], stumps[abs(i)]]
		if i < 0:
			c = [2, c[3], c[4], c[1], c[2]]
		stumps[abs(i) - 1] = m + 1
		stumps[abs(i)] = m
		m += 2
		r.append(c)
	x = [item for subl in r for item in subl[1:]]
	for i in range(len(stumps)):
		if (not stumps[i] in x):
			print("At the moment, all strands of a braid must have at least one crossing.")
			assert(False)
		j = x.index(stumps[i])
		r[int(j / 4)][1 + j % 4] = i 
	return r	



#!/usr/bin/python3

#
#    KrasnerGaussToMyPD.py --- Part of khoca, a knot homology calculator
#
# Copyright (C) 2018 Lukas Lewark <lukas@lewark.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# 2n points on a circle are enumerated from 1 to 2n.
# Each of the points is paired with another one,
# and the connection may be positive or negative.
# This is encoded as a list of integers of even length:
# The element at index i specifies to which point the
# point i is connected; its sign is the sign of the connection.
# To compress, drop half of the list items, namely for each
# connection the higher number. The encoding remains injective.

import sys, re
sys.path.append('.')
import KrasnerGaussToMyPDLib

if len(sys.argv) != 2:
	print("Expecting one argument.")
	sys.exit()
if sys.argv[1] == "-":
	x = sys.stdin.read()
else:
	x = sys.argv[1]
gauss = [int(j) for j in re.findall("-?[0-9]+", x)]
print(str(KrasnerGaussToMyPDLib.KrasnerGaussToMyPDmain(gauss)).replace(" ",""))
